"""
Annnual Conference of the PHM Society 19 - Prognostics Tutorial
Data-driven methods - Example on Gaussian Process Regression

by Corbetta, Matteo
    SGT Inc., NASA Ames Research Center

Summary:
This code presents an example of non-trivial Gaussian Process (GP) regression (GPR) on a one dimensional time series.
Although this example has already been shown in the Scikit-learn libary (see credits), the code has been written using
only numpy and scipy libraries, as well as the matplotlib library for plotting.
The user can therefore follow the code step-by-step to become more familiar with GP, and particularly, GPR.

Notes:
The results from both local and global optimizer have been hard-coded to show all results during the tutorial. However, the code allows the estimation of the hyperparameters
using either differential evolution (DE, global optimization) or the Limited-memory BFGS algorithm (fmin_l_bfgs_b).
Since gradients of kernel functions were not coded, minimization using fmin_l_bfgs_b is likely going to produce poor results because of the complex shape of the log-marginal likelihood hyper-surface.
A better implementation would require to code the gradient of kernels AND log-marginal likelihood w.r.t. the hyperparameters.
DE overcomes the problem, although it requires a LONG time to converge and may still get stuck in a local minima if population size or number of iterations are too small.
DE requires hours to converge on a "normal" laptop.

Dataset:
Data set from Mauna Loa Observatory in Hawaii.
The data consists of the monthly average atmospheric CO_2 concentrations (in parts per million by volume (ppmv))
between 1958 and 2001.

Credits:
To Rasmussen and Williams for the problem formulation, approach and solution:
    Rasmussen CE, Williams CK. Gaussian processes for machine learning. Cambridge, MA: MIT Press; 2006.

To Scikit-learn library for useful tips on the implementation and (a) solution.
    Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011.
    see solution at:
    https://scikit-learn.org/stable/modules/gaussian_process.html

Dataset can be retrieved from the code in the Scikit-learn package:
    Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011.
    https://scikit-learn.org/stable/modules/gaussian_process.html

    Or by visiting NOAA website looking for Mauna Loa dataset.

    The results from the optimization with the Scikit-learn library have been hard-coded to compare them against the optimization from
    differential evolution algorithm.

"""

import numpy as np
import scipy as sp
from scipy import optimize # if using older versions of scipy, the optimize layer can be called by scipy.optimize

# If using Python 2x, the interpreter may raise an error about python installed as a framework, so you need to add "Agg" rendering for matplotlib. Uncomment the following 2 lines to use TkAgg
# import matplotlib as mpl
# mpl.use('TkAgg')

import matplotlib.pyplot as plt

"""
DEFINE KERNELS
  
"""

# Defintion of basic distance functions:
def dist(x, y):
    """
    calculate euclidean distance between x and y
    :param x:       n x 1 array, input vector, doubles
    :param y:       m x 1 array, input vector, doubles
    :return:        n x m array, euclidean distance between x and y
    """
    return abs(np.sum(x, axis=1).reshape(-1, 1) - np.sum(y, axis=1).reshape(1,-1))

def sqDist(x, y):
    """
    calculate squared euclidean distance matrix between two vectors x-y
    :param x:       n x 1 array, input vector, doubles
    :param y:       m x 1 array, input vector, doubles
    :return:        n x m array, squared euclidean distance between x and y
    """
    return np.sum(x ** 2.0, axis=1).reshape(-1, 1) + np.sum(y ** 2.0, axis=1).reshape(1,-1) - 2.0 * np.dot(x, y.T)

# Define kernel functions:
def rbf(x, y, lenscale=1.0, amp=1.0):
    """
    Radial-basis function (RBF) kernel, also called squared exponential:

            amp^2 * \exp( -1/2 * ||x - y||^2/ lenscale^2)

    :param x:               input values, double, (num_datapoints x num_inputDimensions)
    :param y:               output values, double, (num_datapoints x 1)
    :param lenscale:        length scale or wave length of the kernel, double, scalar
    :param amp:             amplitude of the kernel (sometimes called signal variance), double, scalar
    :return:                kernel value
    """
    return amp ** 2.0 * np.exp(-0.5 * sqDist(x, y) / lenscale ** 2.0)  # return covariance matrix

def expSineSq(x, y, lenscale=1.0, period=1.0):
    """
    Squared sine exponential kernel

            \exp( -2 * sin^2( \pi / period * |x-y| ) / lenscale^2 )

    Amplitude has not been included in this kernel since it is always used together with a rbf kernel, which has already the amplitude parameter.
    If used alone, this kernel should be multiplied by an amplitude parameters (see other kernels for reference)

    :param x:               input values, double, (num_datapoints x num_inputDimensions)
    :param y:               output values, double, (num_datapoints x 1)
    :param lenscale:        length scale or wave length of the kernel, double, scalar
    :param period:          periodicity of the kernel, double, scalar
    :return:                kernel value
    """
    return np.exp(-2.0 * (np.sin(np.pi / period * dist(x, y)) / lenscale) ** 2.0)


def ratQuad(x, y, lenscale=1.0, amp=1.0, alpha=1.0):
    """
    Rational quadratic kernel

                    \sigma^2 * (1 + ||x - y|| / (2 * \alpha * l^2))^(-\alpha)


    :param x:               input values, double, (num_datapoints x num_inputDimensions)
    :param y:               output values, double, (num_datapoints x 1)
    :param lenscale:        length scale or wave length of the kernel, double, scalar
    :param amp:             amplitude of the kernel (sometimes called signal variance), double, scalar
    :param alpha:           constant parameter, double, scalar
    :return:                kernel value
    """
    return amp ** 2.0 * (1 + sqDist(x, y) / (2.0 * alpha * lenscale ** 2.0)) ** (-alpha)


def Kernel(x, y, theta):
    """
    Call:

                                K = Kernel(x, y, theta)

    Composite kernel in the form:

                                K = k1 + k2 + k3 + k4

    where: k1 is a radial-basis function (RBF, or squared exponential), k2 is a RBF multiplied by squared sine exponential,
    k3 is a rational quadratic kernel and k4 is a RBF. They represent different features hidden in the dataset.
    The kernel is parametrized by the hyperparameter vector theta, which contains the parameters for each of the sub-kernels
    (k1, k2, etc.) in the CORRECT order. please see the main code for ordering the elements in theta.

    Sub-kernels:

    k1 -> RBF for long term trends

                        k1 = \sigma^2 * \exp( -1/2 * ||x - y||^2/ l^2)

    k2 -> RBF * squared sine exponential for periodicity

                        k2 = \sigma^2 *  \exp( -1/2 * ||x - y||^2/ l^2) * \exp( -2 * sin^2( \pi / T * |x-y| ) / l^2 )

    k3 -> rational quadratic for mid term fluctuations

                        k3 = \sigma^2 * (1 + ||x - y|| / (2 * \alpha * l^2))^(-\alpha)

    k4 -> RBF for short term perturbations

                        k4 = \sigma^2 * \exp( -1/2 * ||x - y||^2/ l^2)

    NB: measurement noise is not embedded into the kernel but rather added during GP regression or when computing
    the log marginal likelihood.

    :param x:           input vector, multi-dimensional array with (num_datapoints x num_input_dimension)
    :param y:           output vector, 1D array (num_datapoints x 1)
    :param theta:       hyperparameter vector
    :return:
    """
    # Kernel 1: long term trend
    k1 = rbf(x, y, lenscale=theta[0], amp=theta[1])

    # Kernel 2: periodicity
    k2 = rbf(x, y, lenscale=theta[2], amp=theta[3]) * expSineSq(x, y, lenscale=theta[4], period=theta[5])

    # Kernel 3: medium term irregularities
    k3 = ratQuad(x, y, lenscale=theta[6], amp=theta[7], alpha=theta[8])

    # kernel 4: short term irregularities
    k4 = rbf(x, y, lenscale=theta[9], amp=theta[10])

    K = k1 + k2 + k3 + k4
    return K


"""
DEFINE GP FUNCTIONS: REGRESSION FUNCTION AND  LOG MARGINAL LIKELIHOOD FUNCTION

"""


def loglikelihoodFunc(theta, *args):
    """
    Call:
        logL = loglikelihoodFun(theta, args)

    log marginal likelihood estimation of the data given the input for squared exponential kernel.
    The parameters are collected within the vector theta:
        theta = (signalVariance, wavelength, measureNoise)

    The squared exponential kernel follows:


        k(x, y) = Sf2 * exp( - 1 / (2 * l^2) * || x - y ||^2 ) + Sn2

    where:
        Sf2 is the amplitude (amp)
        l is the wavelength (wl)
        Sn2 is the measure noise (noise)

    The log marginal likelihood provides the value p(y|X), assuming everything is Gaussian.

    :param theta:       (n,) array of squared exponential parameters: signalVariance, wavelength and measureNoise
    :param args:        input X and output y in a tuple (X, y), necessary to evaluate the log marginal likelihood
    :return:            negative log marginal likelihood given the input (and the parameters).

    """

    # -------- Assign input values ------------ #
    # ls_long, amp_long, ls_period, amp_period, ls_sine, period_sine, ls_med, alpha_med, amp_med, ls_short, amp_short, noiseVar = theta

    # amp, ls, noise = theta  # Extract GP parameters from array theta
    x, y = args  # Extract input and output from array of function arguments

    # -------- Compute kernel ------------ #
    m = x.shape[0]  # get dimension of input space
    K = Kernel(X, X, theta) + theta[-1] * np.eye(m)

    # --- Compute log-marginal likelihood by cholesky decomposition ----- #
    L = np.linalg.cholesky(K + 1e-15 * np.eye(m))  # 1e-15 added to avoid singularities
    alpha = np.linalg.solve(L.T, np.linalg.solve(L, y))
    loglikelihood = -(0.5 * np.dot(y.T, alpha) + np.log(np.diag(L)).sum() + m / 2.0 * np.log(2.0 * np.pi))
    return -loglikelihood


def GPregression(X, y, xtest, theta):
    """
    Call:

        f, covf, logL = GPregression(X, y, xtest, theta)


    Predictions and log-marginal likelihood for Gaussian Process regression.
    Algorithm 2.1, page 19, in "Gaussian Processes for Machine Learning", from Rasmussen and Williams.

    The function computes the predicted mean and covariance of a function fitting the training set {X, y}.
    Noisy observations are assumed to be unbiased and normally distributed, with measure noise variance sn2.
    kerneltype defines the type of kernel. Right now, only squared exponential is implemented.

    The function computes the expected mean and variance using cholesky decomposition, and adding some perturbation
    noise to the covariance matrix of training input for stability.

    :param X:                       m x 1 array, training input
    :param y:                       m x 1 array, training output
    :param xtest:                   n x 1 array, test input
    :param sn2:                     positive float (>=0), measurement noise variance
    :param kernelparams:            p x 1 array, p parameters of the kernel, depending on kernel type. right now only se is implemented, so only wavelength and noise scale parameters can be defined
    :return: f:                     n x 1 array, predicted mean function
    :return: covf:                  n x n array, predicted covariance function
    :return: logp:                  float, log-marginal likelihood of observations y given input data X.
    """

    # ----- Define useful quantities ------------- #
    n = len(xtest)  # number of test input
    m = len(X)  # number of training input

    # normalize mean value
    ymean = np.mean(y)
    y = y - ymean

    # ----- Define Covariance matrix elements using the global kernel -------- #
    K = Kernel(X, X, theta) + theta[-1] * np.eye(m)
    Ks_ = Kernel(xtest, X, theta)
    Kss = Kernel(xtest, xtest, theta)

    # ---------- Cholesky decomposition to compute mean and covariance -------- #
    L = np.linalg.cholesky(K + 1e-15 * np.eye(m))
    alpha = np.linalg.solve(L.T, np.linalg.solve(L, y))
    v = np.linalg.solve(L, Ks_.T)

    # --------- Compute output of the function: mean, cov, and logp ------ #
    f = np.dot(Ks_, alpha)  # mean function prediction
    f = f + ymean  # in case the mean has been normalized
    covf = Kss - np.dot(v.T, v) + theta[-1] * np.eye(n)  # covariance function prediction. for prediction, you need to add the measure noise sn2
    logL = -0.5 * np.dot(y.T, alpha) - sum(np.log(np.diag(L))) - m / 2.0 * np.log(2 * np.pi)  # log-marginal likelihood of observations
    return f, covf, logL


"""
BUILD HYPERPARAMETER OPTIMIZER
"""


def hyperParamOptimizer(input_val, output_val, hyperParamBounds, method='fmin_l_bfgs_b', hyperParam0=None):
    """
    Call:
        hypOpt, logL_val = hyperParamOptimizer(input_val, output_val, hyperParamBounds, method='fmin_l_bfgs_b', hyperParam0=hyp_init)

        hypOpt, logL_val = hyperParamOptimizer(input_val, output_val, hyperParamBounds, method='differential_evolution')

    Description:

    Hyper parameter optimization using a bounded minimization algorithm (fmin_l_bfgs_b) or a global
    optimum algorithm (differential_evolution).

    if using fmin_l_bfgs_b, then an inintial guess of hyperparameter shouold be provided.
    The differential_evolution (DE) algorithm does not need an initial guess, but it is much slower than fmin_l_bfgs_b (up to several hours)

    NOTE:
        fmin_l_bfgs_b currently does not work very well because the gradient of the loglikelihood function is not provided,
        and therefore the algorithm needs to approximate the gradient numerically (see the option approx_grad=True).
        This happens because the kernel (K = k1+k2+k3+k4) is very complex. The algorithm works fine for simpler kernels.

        As a result, the initial guess should already be close to a sub-optimal solution to converge.

        if the log-likelihood function implemented here is enhanced by the gradient computation, the minimization using
        fmin_l_bfgs_b algorithm would be enhanced and would match the performance of the scikit-learn GP package.

        See:
        https://scikit-learn.org/stable/modules/gaussian_process.html
        for more details on GPR regression using the scikit-learn GPR package.

        However, DE is expected to work better than fmin_l_bfgs_b even when the gradient of the loglikelihood
        is computed. This is due to the global optimization nature of DE.

    CREDITS:
        for the concept of optimization using fmin_l_bfgs_b:

        Scikit-learn library:
        https://scikit-learn.org/stable/modules/gaussian_process.html


    :param input_val:                   input data points, double, (num_datapoints x num_inputDimensions)
    :param output_val:                  output values, double, (num_datapoints x 1)
    :param hyperParamBounds:            bounds of hyperparameter, double, array of tuples, (num_parmeters x 2)
    :param method:                      method to fine optimal (or sub-optimal) hyperparameter, string, 'fmin_l_bfgs_b' or 'differential_evolution'
    :param hyperParam0:                 initial guess for hyperparameter optimization, double, (num_parameters x 1)
    :return hypOpt:                     optimized hyperparameters, double, (num_parameters x 1)
    :return logL_val:                   log marginal likelihood value with optimal (or sub-optimal) hyperparameters
    """
    print('Running optimization using ' + method + ' ...')

    if method == 'fmin_l_bfgs_b':
        hypOpt, logL_val, _ = optimize.fmin_l_bfgs_b(loglikelihoodFunc,
                                                                    hyperParam0,                     # initial value of hyperparameter optimization
                                                                    bounds=hyperParamBounds,         # hyperparameter bounds (optional)
                                                                    args=(input_val, output_val),    # arguments to loglikelihoodFun (i.e., constant input). In this case, the constant input is the set of data points (x, y)
                                                                    approx_grad=True,                # approximate gradient of loglikelihood function. If False, loglikelihoodFunc should also provide the gradient. Try to implement it
                                                                    epsilon=1e-8,                    # step size to numerically-approximate the gradient, active only if approx_grad==True
                                                                    factr=1e4,                       # relative tolerance to stop optimization. Absolute tolerance is factr * machine precision
                                                                    maxiter=3500,                    # max number of iterations
                                                                    maxfun=3500,                     # max number of function evaluations
                                                                    maxls=30,                        # maximum number of line search per iteration.
                                                                    disp=1)                          # number of elements to display in terminal during optimization

    elif method == 'differential_evolution':
        de_result = optimize.differential_evolution(loglikelihoodFunc,
                                                    bounds=hyperParamBounds,                # Hyperparameter bounds
                                                    args=(input_val, output_val),           # arguments to loglikelihoodFun (i.e., constant input). In this case, the constant input is the set of data points (x, y)
                                                    disp=True,                              # Display status messages
                                                    polish=True,                            # polishing final solution using fmin_l_bfgs_b algorithm, which can improve minimization slightly
                                                    strategy='best1bin',                    # differential evolution strategy
                                                    maxiter=1500,                           # max number of generations over which the population is evolved
                                                    tol=0.005,                              # relative tolerance for convergence
                                                    mutation=(0.4, 1),                      # mutation constant or differential weight. Used a tuple to take advantage of dithering. The algorithm changes mutation constant randomly between the two values to speed up convergence.
                                                    popsize=150,                            # relative population size, absolute population size is popsize * len(input)
                                                    recombination=0.7)                      # recombination constant, within [0,1]. Larger values allows more mutants to pass to next generation, at the cost of stability.
        hypOpt = de_result.x
        logL_val = de_result.fun

    else:
        print('method not recognized')

    print('************************************************')
    print('Optimization completed.')
    print('Parameters:')
    print(hypOpt)
    print(' ')
    print('Log marginal likelihood:')
    print(logL_val)
    print(' ')
    print('************************************************')

    return hypOpt, logL_val


def assignHyperparameterValues(assignTheta='guess'):
    """
    Call:

        theta = assignHyperparameterValues(assignTheta='guess')
        theta = assignHyperparameterValues(assignTheta='optimal')
        theta = assignHyperparameterValues(assignTheta='scikitValues')

    Description:

    assign hyperparameter value to hyperparameter vector theta

    the function assign hyperparameter values to vector theta for the regression or optimization.

    Input options: 'guess', 'optimal', 'scikitValues'


    if assignTheta = 'guess', an initial (reasonably good) guess is assigned to theta for optimization using
    fmin_l_bfgs_b algorithm

    if assignTheta = 'optimal', the values computed with a run of differential_evolution algorithm are assigned to theta

    if assignTheta = 'scikitValues', the values posted in the Scikit-learn page https://scikit-learn.org/stable/modules/gaussian_process.html
    are assigned to theta.
    These values can be compared against the 'optimal' values from differential evolutions to show how the optimal solution
    behaves against a sub-optimal one.




    :param assignTheta:         string, assignment type, see description for more details
    :return theta:              nparameters x 1 array of GP hyperparameters
    """

    if assignTheta == 'guess':
        # Initial guess for the hyperparameters. It does not produce good extrapolation capabilities but it can be
        # used as initial guess of the fmin_l_bfgs_b optimizer

        theta1 = np.array([50.0, 50.0])  # wave length and amplitude of k1, squared exponential for long term trends
        theta2 = np.array(
            [10.0, 10.0, 10.0, 1.0])  # wave length and amplitude of k2, squared exponential for periodic term, and wave length and period of k2, squared sine exponential for periodic term
        theta3 = np.array([1.0, 1.0, 1.0])  # wave length, amplitude and coefficient of k3, rational quadratic for medium term trends
        theta4 = np.array([0.1, 0.1])  # wave length and amplitude of k4, squared exponential for short term fluctuations
        theta5 = 0.01  # variance of noise term

    elif assignTheta == 'optimal':
        # hyperparameter values obtained after a long run of differential_evolution (the optimization lasted for a few hours).
        # it is not guaranteed that the values below are the true global optima given the complexity of the log marginal likelihood.
        # However, they generate good extrapolation capabilities.

        theta1 = np.array([1.13014358e+02, 3.10438494e+02])
        theta2 = np.array([1.11582307e+02, 2.63909877e+00, 1.47835402e+00, 9.99529318e-01])
        theta3 = np.array([1.47496439e+00, 9.23866714e-01, 1.93912266e-01])
        theta4 = np.array([1.20838358e-01, 1.76876769e-01])
        theta5 = 3.78704721e-02

    elif assignTheta == 'scikitValues':
        # try prediction with suboptimal theta, based on fmin_l_bfgs_b values from Scikit-learn package (see https://scikit-learn.org/stable/auto_examples/gaussian_process/plot_gpr_co2.html)
        # The hyperparameter optimization using the Scikit-learn GPR package produces better results because the gradient of the log-likelihood function
        # is coded into the library and it is not numerically approximated by the optimizer.

        theta1 = np.array([41.8, 34.4])
        theta2 = np.array([180, 3.27, 1.44, 1.0])
        theta3 = np.array([0.957, 0.446, 17.7])
        theta4 = np.array([0.138, 0.197])
        theta5 = np.array([0.0336])

    else:
        print('Assignment type not recognized')

    # Collect all parameters in a 1D array to pass it to GPregression and likelihood function
    return np.array([theta1[0], theta1[1], theta2[0], theta2[1], theta2[2], theta2[3], theta3[0], theta3[1], theta3[2], theta4[0], theta4[1], theta5])


if __name__ == '__main__':

    # fix random seed for repeatability
    np.random.seed(11223344)
    
    """
    LOAD DATASET
    
    """
    data = np.loadtxt('CO2data.txt', dtype=float, comments='#', skiprows=1)

    # Assign input and output data points to X and y variables
    X = data[:, 2].reshape(-1, 1)  # [decimal years], GP input, recorded
    y = data[:, 4].reshape(-1, 1)  # [ppm], CO_2 concentration, observed

    """
    DEFINE PREDICTION PARAMETERS
    
    """

    prog_horizon = 30.0  # [decimal years], prognostic horizon
    xtest = np.linspace(min(X), max(X) + prog_horizon, num=1000).reshape(-1, 1)  # [decimal years], input vector for testing
    idx_prog = np.argmin(abs(xtest-max(X)))     # index identifying the prognosis step: when x becomes larger than the largest training point

    optimize_params = False
    if optimize_params:
        optimization_type = 'fmin_l_bfgs_b'  # select between 'fmin_l_bfgs_b' or 'differential_evolution'
    else:
        assignThetaValue = 'optimal'  # select among 'guess', 'otpimal', 'scikitValues'

    """
    PARAMETER OPTIMIZATION
    
    
    Optimize model parameters using either bound constrained optimization algorithm fmin_l_bfgs_b
    or using differential evolution.
    
    * fmin_l_bfgs_b gets stuck in local minima, especially because the gradient of the loglikelihood function is not provided
        provinding the explicit gradient formulation of the loglikelihood function (which needs the gradient of the kernel w.r.t. all model parameters
        would help finding better suboptimal points.
    
    * differential_evolution is more robust but needs a lot of time to run (hours).
    
    """

    if optimize_params:
        # Define hyperparameter vector and corresponding bounds for optimization
        hyp_init = assignHyperparameterValues(assignTheta='guess')
        hyp_bounds = [(1.0, 1000.0), (1.0, 1000.0),
                      (1.0, 1000.0), (1.0, 1000.0),
                      (0.01, 500.0), (0.01, 500.0),
                      (0.01, 200.0), (0.01, 200.0), (0.01, 200.0),
                      (0.01, 100.0), (0.01, 100.0),
                      (0.001, 10.0), ]

        hypOpt, logL_val = hyperParamOptimizer(X, y, hyperParamBounds=hyp_bounds, method=optimization_type, hyperParam0=hyp_init)

        theta = hypOpt

    else:
        theta = assignHyperparameterValues(assignTheta=assignThetaValue)

    """
    GAUSSIAN PROCESS REGRESSION
    
    Run the GP regression using the hyperparameters in theta
    
    
    """
    # ------------ Run GP with hyperparameters selected in PARAMETER OPTIMIZATION section ------------------- #
    f, covf, logp = GPregression(X, y, xtest, theta)
    y_pred = f.reshape((-1,))
    y_std = np.sqrt(np.diag(covf)).reshape((-1,))


    # # ------------ Run GP with hyperparameters optimized with Scikit-learn package ---------- #
    # theta = assignHyperparameterValues(assignTheta='scikitValues')
    # f, covf, logp = GPregression(X, y, xtest, theta)
    # y_pred_skl = f.reshape((-1,))
    # y_std_skl  = np.sqrt(np.diag(covf)).reshape((-1,))


    """
    PLOTTING
    
    Compare observed data points and forecast
    
    Observed CO2 concentration in April, 2018 and April, 2019: 410.24 ppm and 413.32 ppm
    Data retrieved from: 
    https://www.esrl.noaa.gov/gmd/ccgg/trends/
    
    """

    plt.figure()
    plt.plot(X, y, 'b-', linewidth=2, label=r'$\mathrm{CO}_2$')
    plt.xlabel('year, -', fontsize=14)
    plt.ylabel(r'concentration, ppm', fontsize=14)
    plt.legend(fontsize=14, fancybox=True, shadow=True)
    plt.savefig('CO2_dataPoints.pdf', dpi=300)

    # measurements from April 2018 and 2019
    xobs = np.array([2018.292, 2019.292])
    yobs = np.array([410.24, 413.32])

    # plot predictions from 2001 on only
    plt.figure(figsize=(8, 7))
    plt.plot(X, y, 'b-', linewidth=2, label='observations')
    plt.plot(xtest[idx_prog:], y_pred[idx_prog:], 'r--', linewidth=2.0, label='prediction')
    # plt.plot(xobs, yobs, 'ko', markersize=12, label='CO$_2$ in Apr 18-19')
    # plt.plot(xobs, yobs, 'ko', markersize=12, label='_nolegend_')
    plt.fill_between(xtest[idx_prog:].reshape((-1,)), y_pred[idx_prog:] - y_std[idx_prog:], y_pred[idx_prog:] + y_std[idx_prog:], color='r', alpha=0.2, label=r'$\pm \sigma$')
    plt.xlabel('time, years', fontsize=14)
    plt.ylabel(r'concentration, ppm', fontsize=14)
    plt.legend(loc='upper left', fontsize=14, fancybox=True, shadow=True)
    plt.savefig('CO2_prediction.pdf', dpi=300)

    # plt.figure(figsize=(8,7))
    # plt.plot(X, y, 'b', linewidth=2, label='observations')
    # plt.plot(xtest[idx_prog:], y_pred[idx_prog:], 'r--', label='prediction')
    # plt.plot(xtest[idx_prog:], y_pred_skl[idx_prog:], 'k--', label='prediction from Scikit-learn')
    # # plt.plot(xobs, yobs, 'ko', markersize=12, label='CO$_2$ in Apr 18-19')
    # # plt.plot(xobs, yobs, 'ko', markersize=12, label='_nolegend_')
    # plt.fill_between(xtest[idx_prog:].reshape((-1,)), y_pred[idx_prog:] - y_std[idx_prog:], y_pred[idx_prog:] + y_std[idx_prog:], color='r', alpha=0.2, label=r'$\pm \sigma$')
    # plt.fill_between(xtest[idx_prog:].reshape((-1,)), y_pred_skl[idx_prog:] - y_std_skl[idx_prog:], y_pred_skl[idx_prog:] + y_std_skl[idx_prog:], color='k', alpha=0.2, label=r'$\pm \sigma$ from Scikit-learn')
    # plt.xlabel('time, years', fontsize=14)
    # plt.ylabel(r'concentration, ppm', fontsize=14)
    # plt.legend(loc='lower right', fontsize=14, fancybox=True, shadow=True)
    # plt.savefig('CO2_predictionCompare.pdf', dpi=300)

    # fig = plt.figure(figsize=(8, 7))
    # ax = fig.add_subplot(111)
    # ax.plot(X, y, 'b-', linewidth=2, label=r'CO$_2$ at Mauna Loa, Hawaii')
    # ax.plot(xtest[idx_prog:], y_pred[idx_prog:], 'r--', label='prediction')
    # ax.plot(xtest[idx_prog:], y_pred_skl[idx_prog:], 'k--', label='Scikit-learn prediction')
    # ax.plot(xobs, yobs, 'o', color='coral', markersize=12, label='_nolegend_')
    # ax.fill_between(xtest[idx_prog:].reshape((-1,)), y_pred[idx_prog:] - y_std[idx_prog:], y_pred[idx_prog:] + y_std[idx_prog:], color='r', alpha=0.2, label=r'$\pm \sigma$')
    # ax.fill_between(xtest[idx_prog:].reshape((-1,)), y_pred_skl[idx_prog:] - y_std_skl[idx_prog:], y_pred_skl[idx_prog:] + y_std_skl[idx_prog:], color='k', alpha=0.2,
    #                  label=r'$\pm \sigma$ Scikit-learn')
    # props = dict(boxstyle='round', facecolor='w', alpha=0.5)
    #
    # # pointer to Apr 2018
    # ax.arrow(2001, 410.0, 14.5, 0.5, head_width=2.0, head_length=1.5, color='k')
    # textstr = r'Apr $^\prime$18'
    # ax.text(0.45, 0.8, textstr, transform=ax.transAxes, fontsize=14,
    #         verticalalignment='top', bbox=props)
    #
    # # pointer to Apr 2019
    # ax.arrow(2001, 430.0, 16, -15, head_width=2.0, head_length=1.5, color='k')
    # textstr = r'Apr $^\prime$19'
    # ax.text(0.45, 0.97, textstr, transform=ax.transAxes, fontsize=14,
    #         verticalalignment='top', bbox=props)
    #
    # plt.xlabel('time, years', fontsize=14)
    # plt.ylabel(r'concentration, ppm', fontsize=14)
    # plt.legend(loc='lower right', fontsize=14, fancybox=True, shadow=True)
    # plt.savefig('CO2_predictionCompare_w2018obs.pdf', dpi=300)

    plt.show()
